/* DWARF DW_FORM_block* expression evaluation.

   Copyright (C) 2007 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "defs.h"
#include "dwarf2block.h"
#include "gdbcore.h"
#include "dwarf2expr.h"
#include "exceptions.h"
#include "frame.h"
#include "regcache.h"
#include "value.h"
#include "block.h"
#include "gdb_assert.h"
#include "dwarf2loc.h"

/* This is the baton used when performing dwarf2 DW_BLOCK evaluation.  */
struct dwarf_block_baton
{
  CORE_ADDR address;
};

/* Read memory at ADDR (length LEN) into BUF.  */

static void
dwarf_block_read_mem (void *baton, gdb_byte *buf, CORE_ADDR addr, size_t len)
{
  read_memory (addr, buf, len);
}

static CORE_ADDR
dwarf_block_object_address (void *baton)
{
  struct dwarf_block_baton *debaton = baton;

  /* The message is suppressed in DWARF_BLOCK_EXEC.  */
  if (debaton->address == 0)
    error (_("Cannot resolve DW_OP_push_object_address for a missing object"));

  return debaton->address;
}

/* A copy from dwarf2-frame.c:read_reg() but without one unwind.  */

static CORE_ADDR
dwarf_block_read_reg (void *baton, int reg)
{
  struct frame_info *frame = get_selected_frame
		       (_("Unsupported operation for DW_FORM_block: read_reg"));
  struct gdbarch *gdbarch = get_frame_arch (frame);
  int regnum;
  gdb_byte *buf;

  regnum = gdbarch_dwarf2_reg_to_regnum (gdbarch, reg);

  buf = alloca (register_size (gdbarch, regnum));
  get_frame_register (frame, regnum, buf);

  /* Convert the register to an integer.  This returns a LONGEST
     rather than a CORE_ADDR, but unpack_pointer does the same thing
     under the covers, and this makes more sense for non-pointer
     registers.  Maybe read_reg and the associated interfaces should
     deal with "struct value" instead of CORE_ADDR.  */
  return unpack_long (register_type (gdbarch, regnum), buf);
}

/* A copy from dwarf2loc.c:dwarf_expr_frame_base().  */

static void
dwarf_block_get_frame_base (void *baton, gdb_byte **start, size_t *length)
{
  struct frame_info *frame = get_selected_frame
		       (_("Unsupported operation for DW_FORM_block: read_reg"));
  /* FIXME: cagney/2003-03-26: This code should be using
     get_frame_base_address(), and then implement a dwarf2 specific
     this_base method.  */
  struct symbol *framefunc;

  framefunc = block_function (get_frame_block (frame, NULL));

  /* If we found a frame-relative symbol then it was certainly within
     some function associated with a frame. If we can't find the frame,
     something has gone wrong.  */
  gdb_assert (framefunc != NULL);

  if (SYMBOL_LOCATION_BATON (framefunc) == NULL)
    *start = NULL;
  else if (SYMBOL_OPS (framefunc) == &dwarf2_loclist_funcs)
    {
      struct dwarf2_loclist_baton *symbaton;

      symbaton = SYMBOL_LOCATION_BATON (framefunc);
      *start = find_location_expression (symbaton, length,
					 get_frame_address_in_block (frame));
    }
  else
    {
      struct dwarf2_locexpr_baton *symbaton;

      symbaton = SYMBOL_LOCATION_BATON (framefunc);
      *length = symbaton->size;
      *start = symbaton->data;
    }

  if (*start == NULL)
    error (_("%s: Could not find the frame base for \"%s\"."),
	   "dwarf_block_get_frame_base", SYMBOL_NATURAL_NAME (framefunc));
}

static CORE_ADDR
dwarf_block_get_tls_address (void *baton, CORE_ADDR offset)
{
  error (_("Unsupported operation for DW_FORM_block*: %s"), "get_tls_address");
  return 0;
}

static CORE_ADDR dwarf_block_exec_core (struct dwarf_block *dwarf_block,
					CORE_ADDR address)
{
  struct dwarf_expr_context *ctx;
  struct dwarf_block_baton baton;
  struct cleanup *back_to;
  CORE_ADDR retval;

  back_to = make_cleanup (null_cleanup, 0);

  baton.address = address;

  ctx = new_dwarf_expr_context ();
  back_to = make_cleanup ((make_cleanup_ftype *) free_dwarf_expr_context, ctx);
  ctx->baton = &baton;
  ctx->read_mem = dwarf_block_read_mem;
  ctx->get_object_address = dwarf_block_object_address;
  ctx->read_reg = dwarf_block_read_reg;
  ctx->get_frame_base = dwarf_block_get_frame_base;
  ctx->get_tls_address = dwarf_block_get_tls_address;

  dwarf_expr_eval (ctx, dwarf_block->data, dwarf_block->size);

  if (ctx->num_pieces > 0)
    error (_("DW_OP_piece is an unsupported result for DW_FORM_block*"));
  if (ctx->in_reg)
    error (_("DW_OP_reg* is an unsupported result for DW_FORM_block*"));

  retval = dwarf_expr_fetch (ctx, 0);

  do_cleanups (back_to);

  return retval;
}

static CORE_ADDR object_address;

static void
object_address_cleanup (void *prev_save_voidp)
{
  CORE_ADDR *prev_save = prev_save_voidp;

  object_address = *prev_save;
  xfree (prev_save);
}

void
object_address_set (CORE_ADDR address)
{
  CORE_ADDR *prev_save;

  prev_save = xmalloc (sizeof *prev_save);
  *prev_save = object_address;
  make_cleanup (object_address_cleanup, prev_save);

  object_address = address;
}

CORE_ADDR
object_address_get (void)
{
  return object_address;
}

CORE_ADDR dwarf_block_exec (struct dwarf_block *dwarf_block)
{
  volatile struct gdb_exception e;
  volatile CORE_ADDR retval = 0;

  TRY_CATCH (e, RETURN_MASK_ERROR)
    {
      retval = dwarf_block_exec_core (dwarf_block, object_address);
    }
  /* CATCH_ERRORS would print the possible error message of
     DWARF_BLOCK_OBJECT_ADDRESS.  Sometimes it is valid as CHECK_TYPEDEF is
     a very common call even if we still do not know any variable instance of
     that type.  We cannot fill in the right TYPE_LENGTH at that time.  */
  if (e.reason < 0)
    return 0;

  return retval;
}
